/*-
 * Copyright Sergey Kosyakov ks@itp.ac.ru 1999
 * Copyright (C)2002..2020 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#)Copyright Sergey Kosyakov ks@itp.ac.ru 1999"
#ident "@(#)Copyright (C)2002..2020 @BABOLO http://www.babolo.ru/"
#ident "$Id: secur.c,v 1.22 2020/12/05 17:03:19 babolo Exp $"

#define BLIN_COMPAT  3

#include <openssl/blowfish.h>
#include <openssl/rc5.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sysexits.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <md5.h>
#include <netinet/in.h>
#include <openssl/rsa.h>
#include <openssl/err.h>
#ifndef NO_IDEA
# include <openssl/idea.h>
#endif
#include <babolo/BLINflag.h>
#include "tund.h"

static u_char    iv_buf[8];
static u_char    ivec_ini[8] = {"kjsLd5Gh"};
extern u_char    packetBuf[TUND_PKT_MAXSZ];
static u_char    cipherBuf[TUND_PKT_MAXSZ];
static u_char    sKey[16];
static RSA      *rsa_priv = NULL;
static tund_seq *hostsKey;
static RSA      *rsa_pub;

typedef struct {
    char *host;
    RSA  *rsa;
} HostKey;

static void *
must_malloc(tund_conf *cnf, int size) {
#   define blin_internal_flags (cnf->flags)
    void *ret;

    ret = malloc(size);
    if  (!ret) {
        ifBLIN_QX0("Can not allocate %d bytes in must_malloc", size);
        exit(1);
    }
    return(ret);
#   undef blin_internal_flags
}

static void
tund_set_key(tund_conf *cnf, int tun_ind, int remote) {
#   define blin_internal_flags (cnf->flags)
    RC5_32_KEY           *rc5_key;
    BF_KEY               *bf_key;
    int                   cipher;
    u_char               *data;
    tund_tunnel          *tun;
    void                 *key;
#   ifndef NO_IDEA
       IDEA_KEY_SCHEDULE *idea_key1;
       IDEA_KEY_SCHEDULE *idea_key2;
#   endif

    tun = (tund_tunnel *)cnf->stun->buf[tun_ind];
    cipher = (tun->flags) & TUND_CIPHER_MASK;
    if  (remote) {
        if  (tun->remoteCipherKey) free(tun->remoteCipherKey);
        tun->remoteCipherKey = NULL;
        data = tun->remote_key;
    } else {
        if  (tun->localCipherKey) free(tun->localCipherKey);
        tun->localCipherKey = NULL;
        data = tun->local_key;
    }
    switch(cipher) {
    case TUND_CIPHER_BF:
        bf_key = must_malloc(cnf, sizeof(BF_KEY));
        BF_set_key(bf_key, 16, data);
        key = bf_key;
        break;
#ifndef NO_IDEA
    case TUND_CIPHER_IDEA:
        idea_key1 = must_malloc(cnf, sizeof(IDEA_KEY_SCHEDULE));
        idea_set_encrypt_key(data, idea_key1);
        if  (remote) {
            idea_key2 = must_malloc(cnf, sizeof(IDEA_KEY_SCHEDULE));
            idea_set_decrypt_key(idea_key1, idea_key2);
            free(idea_key1);
            idea_key1 = idea_key2;
        }
        key = idea_key1;
        break;
#endif
    case TUND_CIPHER_RC5:
        rc5_key = must_malloc(cnf, sizeof(RC5_32_KEY));
        RC5_32_set_key(rc5_key, 16, data, RC5_8_ROUNDS);
        key = rc5_key;
        break;
    default:
        ifBLIN_QX0("Unknown cipher 0x%X", cipher);
        exit(1);
    }
    if  (remote) tun->remoteCipherKey = key; else tun->localCipherKey = key;
    ifBLIN_QX5( "New %s key for cipher 0x%X for tunnel %u"
              , remote ? "remote" : "local"
              , cipher
              , tun->tunum
              );
#   undef blin_internal_flags
}

static int
tund_cipher_enc(tund_conf *cnf, u_char *in, u_char *out, int len, int tun_ind) {
#   define blin_internal_flags (cnf->flags)
    tund_tunnel *tun = (tund_tunnel *)cnf->stun->buf[tun_ind];
    int          cipher;
    BF_KEY      *bf_key;
    RC5_32_KEY  *rc5_key;
    int          length;
#   ifndef NO_IDEA
        IDEA_KEY_SCHEDULE *idea_key;
#   endif

    length = ((len / 8) + 1) * 8;
    cipher = (tun->flags) & TUND_CIPHER_MASK;
    switch(cipher) {
    case TUND_CIPHER_BF:
        bf_key = tun->localCipherKey;
        bcopy(ivec_ini, iv_buf, 8);
        BF_cbc_encrypt(in, out, length, bf_key, iv_buf, BF_ENCRYPT);
        return(length);
#ifndef NO_IDEA
    case TUND_CIPHER_IDEA:
        idea_key = tun->localCipherKey;
        bcopy(ivec_ini, iv_buf, 8);
        idea_cbc_encrypt(in, out, length, idea_key, iv_buf, IDEA_ENCRYPT);
        return(length);
#endif
    case TUND_CIPHER_RC5:
        rc5_key = tun->localCipherKey;
        bcopy(ivec_ini, iv_buf, 8);
        RC5_32_cbc_encrypt(in, out, length, rc5_key, iv_buf, RC5_ENCRYPT);
        return(length);
    default:
        ifBLIN_QX0("Unknown cipher 0x%X", cipher);
        exit(1);
    }
    return(-1);
#   undef blin_internal_flags
}

static void
tund_cipher_dec(tund_conf *cnf, u_char *in, u_char *out, int len, int tun_ind) {
#   define blin_internal_flags (cnf->flags)
#ifndef NO_IDEA
    IDEA_KEY_SCHEDULE *idea_key;
#endif
    tund_tunnel *tun = (tund_tunnel *)cnf->stun->buf[tun_ind];
    RC5_32_KEY *rc5_key;
    BF_KEY *bf_key;
    int cipher;
    int length;

    length = len;
    cipher = (tun->flags) & TUND_CIPHER_MASK;
    switch(cipher) {
    case TUND_CIPHER_BF:
        bf_key = tun->remoteCipherKey;
        bcopy(ivec_ini, iv_buf, 8);
        BF_cbc_encrypt(in, out, length, bf_key, iv_buf, BF_DECRYPT);
        break;
#ifndef NO_IDEA
    case TUND_CIPHER_IDEA:
        idea_key = tun->remoteCipherKey;
        bcopy(ivec_ini, iv_buf, 8);
        idea_cbc_encrypt(in, out, length, idea_key, iv_buf, IDEA_DECRYPT);
        break;
#endif
    case TUND_CIPHER_RC5:
        rc5_key = tun->remoteCipherKey;
        bcopy(ivec_ini, iv_buf, 8);
        RC5_32_cbc_encrypt(in, out, length, rc5_key, iv_buf, RC5_DECRYPT);
        break;
    default:
        ifBLIN_QX0("Unknown cipher 0x%X", cipher);
        exit(1);
    }
#   undef blin_internal_flags
}

int
tund_init_secur(tund_conf *cnf) {                                        /* Initialization procedure */
#   define blin_internal_flags (cnf->flags & BLIN_MASK)
    char           hostname[128];
    char           str[8192];
    u_char         key[8192];
    RSA           *rsa;
    int            ret;
    int            len;
    int            ex = EX_OK;
    HostKey       *hk;
    unsigned int   n;
    const u_char  *p;
    FILE          *f;

    /* Read own RSA private key */
    if  (!(f = fopen(cnf->rsak, "r"))) {
        ifBLIN_QW0("Can not open %s", cnf->rsak);
        ex = EX_IOERR;
        goto out;
    }
    bzero(key, 8192);
    len = 0;
    do {
        ret = fscanf(f, "%2x", &n);
        if  (ret != 1) break;
        key[len] = n;
        len++;
    } while(len < 8190);
    ifBLIN_QX3("Read %d bytes as RSA private key", len);
    fclose(f);
    p = key;
    rsa = d2i_RSAPrivateKey(&rsa_priv, &p, (long)len);
    if  (!rsa || !rsa_priv) {
        ifBLIN_QX0("Error while d2i_RSAPrivateKey");
        ex = EX_CONFIG;
        goto out;
    }
    /* Now scan other hosts RSA public key database */
    if  (!(f = fopen(cnf->rsap, "r"))) {
        ifBLIN_QX0("Can not open %s", cnf->rsap);
        ex = EX_IOERR;
        goto out;
    }
    hostsKey = tund_seq_new(cnf); 
    do {
        bzero(key, 8192);
        bzero(str, 8192);
        bzero(hostname, 128);
        if  (0 >= (ret = fscanf(f, "%s %s", hostname, str))) break;
        ifBLIN_QX5("Host [%s] key [%s]", hostname, str);
        for (len = 0; len * 2 < strlen(str); len++) {
            if  (1 != (ret = sscanf(str + len * 2, "%2x", &n))) {
                ifBLIN_QW0("Corrupted %s", cnf->rsap);
                ex = EX_IOERR;
                goto out;
            }
            key[len] = n;
        }
        ifBLIN_QX3("Read %d bytes as RSA public key", len);
        p = key;
        rsa_pub = NULL;
        rsa = d2i_RSAPublicKey(&rsa_pub, &p, (long)len);
        if  (!rsa || !rsa_pub) {
            ifBLIN_QX0("Error while d2i_RSAPublicKey");
            ex = EX_CONFIG;
            goto out;
        }
        hk = (HostKey *)malloc(sizeof(HostKey));
        hk->host = strdup(hostname);
        hk->rsa = rsa_pub;
        tund_seq_append(cnf, hostsKey, (void *)hk);
    } while(1);
out:
    return(ex);
#   undef blin_internal_flags
}

int
tund_data_enc(tund_conf *cnf, u_char *buf, int len, int tun_ind) {
#   define blin_internal_flags (cnf->flags & BLIN_MASK)
    tund_tunnel *tun = (tund_tunnel *)cnf->stun->buf[tun_ind];
    int en_len;

    if  (  !(tun->flags & TUND_CIPHER_MASK)
        || !(tun->flags & TUND_F_RPWD)
        || (tun->state != TUND_S_UP)
        ) return(-1);
    ifBLIN_QX5("Request for encryption of %d bytes t=%d", len, tun_ind);
    if  ((en_len = tund_cipher_enc(cnf, buf, cipherBuf, len, tun_ind)) < 0) return(-1);
    bcopy(cipherBuf, buf, en_len);
    return(en_len);
#   undef blin_internal_flags
}

void
tund_data_dec(tund_conf *cnf, u_char *buf, int len, int tun_ind) {
#   define blin_internal_flags (cnf->flags & BLIN_MASK)
    ifBLIN_QX5("Request for decryption of %d bytes t=%d", len, tun_ind);
    tund_cipher_dec(cnf, buf, cipherBuf, len, tun_ind);
    bcopy(cipherBuf, buf, len);
#   undef blin_internal_flags
}

int
rsa_encrypt(tund_conf *cnf, u_char *buf, int len) {
#   define blin_internal_flags (cnf->flags & BLIN_MASK)
    static u_char rsa_str[1024];
    int ret;

    ret = RSA_private_encrypt(len, buf, rsa_str, rsa_priv, RSA_PKCS1_PADDING);
    if  (ret <= 0) {
        ifBLIN_QX0( "rsa_encrypt(0x%x %d) returns %d: %s"
                  , buf
                  , len
                  , ret
                  , ERR_error_string(ERR_peek_last_error(), NULL)
                  );
        exit(1);
    }
    bcopy(rsa_str, buf, ret);
    ifBLIN_QX5("RSA encrypted %d bytes into %d bytes", len, ret);
    return(ret);
#   undef blin_internal_flags
}

int
rsa_decrypt(tund_conf *cnf, tund_tunnel *tun, u_char *buf, int len) {
#   define blin_internal_flags (cnf->flags & BLIN_MASK)
    static u_char rsa_str[1024];
    int ret;

    ifBLIN_QX5("Request for RSA decrypt of %d bytes", len);
    ret = RSA_public_decrypt(len, buf, rsa_str, tun->r_key, RSA_PKCS1_PADDING);
    if  (ret <= 0) {
        ifBLIN_QX3( "rsa_decrypt(0x%x %d) returns %d: %s"
                  , buf
                  , len
                  , ret
                  , ERR_error_string(ERR_peek_last_error(), NULL)
                  );
        return(ret);
    }
    ifBLIN_QX5("RSA decrypted %d bytes into %d bytes", len, ret);
    bcopy(rsa_str, buf, ret);
    return(ret);
#   undef blin_internal_flags
}

static int
control_encrypt(tund_conf *cnf, tund_CP *cp) {
#   define blin_internal_flags (cnf->flags & BLIN_MASK)
    int len;

    cp->c.rn = tund_get_rand(cnf);
    len = rsa_encrypt( cnf
                     , ((u_char *)cp) + sizeof(tund_pkthead)
                     , cp->c.body_len + sizeof(tund_ctrlhead)
                     );
    return(len + sizeof(tund_pkthead));
#   undef blin_internal_flags
}

int
control_decrypt(tund_conf *cnf, tund_CP *cp, tund_tunnel *tun, int size) {
#   define blin_internal_flags (cnf->flags & BLIN_MASK)
    int ret;

    ret = rsa_decrypt(cnf, tun, ((u_char *)cp) + sizeof(tund_pkthead), size-sizeof(tund_pkthead));
    return(ret);
#   undef blin_internal_flags
}

static void
send_control(tund_conf *cnf, int tun_ind) {
#   define blin_internal_flags (cnf->flags & BLIN_MASK)
    tund_CP           *control;
    struct sockaddr_in saddr;
    tund_tunnel       *tun;
    int                len;

    tun = (tund_tunnel *)cnf->stun->buf[tun_ind];
    control = (tund_CP *)packetBuf;
    len = control_encrypt(cnf, control);
    bcopy(&(tun->to), &saddr, sizeof(struct sockaddr_in));
    saddr.sin_port = cnf->s_port;
    if  (!!tund_ks_send( cnf
                       , tun_ind
                       , cnf->isock
                       , packetBuf
                       , len
                       , (struct sockaddr *)(&saddr)
                       , sizeof(struct sockaddr_in)
        )              ) {
        ifBLIN_QW2("");
        if  (EIO == errno) exit(EX_IOERR);
    }
#   undef blin_internal_flags
}

static int
reset_remote_tunnel(tund_conf *cnf, int tun_ind) {
#   define blin_internal_flags (cnf->flags & BLIN_MASK)
    tund_CP     *control;
    int          ex = EX_OK;
    tund_tunnel *tun;

    tun = (tund_tunnel *)cnf->stun->buf[tun_ind];
    if  (tun->state != TUND_S_RSTAW) {
        ifBLIN_QX1("tunnel %d state %d <> %d", tun_ind, tun->state, TUND_S_RSTAW);
        goto out;
    }
    control = (tund_CP *)packetBuf;
    control->h.ver = TUND_PROTO_VER;
    control->h.flags = TUND_PPF_CNT;
    control->h.tunum = tun->tunum;
    control->c.mt = TUND_CMT_RESET;
    control->c.body_len = 0;
    send_control(cnf, tun_ind);
    tun->state = TUND_S_RSTAW;
    tund_add_expired_operation(cnf, tun_ind, 10, reset_remote_tunnel);
    ifBLIN_QX5("Reset tunnel %u requested", tun->tunum);
out:
    return(ex);
#   undef blin_internal_flags
}

int
tund_start_sectun(tund_conf *cnf, int tun_ind) {
#   define blin_internal_flags (cnf->flags & BLIN_MASK)
    int          ex = EX_OK;
    tund_tunnel *tun;
    HostKey     *hk;
    int          i;

    tun = (tund_tunnel *)cnf->stun->buf[tun_ind];
    /* First find remote end RSA public key */
    for (i = 0; i < hostsKey->length; i++) {
        hk = (HostKey *)hostsKey->buf[i];
        if  (strcmp(hk->host, tun->remoteName) == 0) {
            tun->r_key = hk->rsa;
            ifBLIN_QX5( "Tun %u remote h=%s RSA PKEY at 0x%x"
                      , tun->tunum
                      , tun->remoteName
                      , tun->r_key
                      );
            break;
    }   }
    if  (!tun->r_key) {
        ifBLIN_QX0("Can not find RSA PKEY for tun %u", tun->tunum);
        ex = EX_USAGE;
        goto out;
    }
    tun->state = TUND_S_RSTAW;                                         /* First reset tunnel on peer */
    reset_remote_tunnel(cnf, tun_ind);
out:
    return(ex);
#   undef blin_internal_flags
}

static int
send_new_password(tund_conf *cnf, int tun_ind) {
#   define blin_internal_flags (cnf->flags & BLIN_MASK)
    tund_CP     *control;
    int          ex = EX_OK;
    tund_tunnel *tun;

    tun = (tund_tunnel *)cnf->stun->buf[tun_ind];
    if  (tun->state != TUND_S_SPAW) {
        ex = EX_SOFTWARE;
        goto out;
    }
    control = (tund_CP *)packetBuf;
    control->h.ver = TUND_PROTO_VER;
    control->h.flags = TUND_PPF_CNT;
    control->h.tunum = tun->tunum;
    control->c.mt = TUND_CMT_SPWD;
    control->c.body_len = 16;
    bcopy(tun->local_key, control->body, 16);
    send_control(cnf, tun_ind);
    tun->state = TUND_S_SPAW;
    tund_add_expired_operation(cnf, tun_ind, 10, send_new_password);
    ifBLIN_QX5("New password for tunnel %u sent", tun->tunum);
out:
    return(ex);
#   undef blin_internal_flags
}

static int
accept_new_password(tund_conf *cnf, int tun_ind) {
#   define blin_internal_flags (cnf->flags & BLIN_MASK)
    tund_CP     *control;
    int          ex = EX_OK;
    tund_tunnel *tun;

    tun = (tund_tunnel *)cnf->stun->buf[tun_ind];
    control = (tund_CP *) packetBuf;
    bcopy(control->body, tun->remote_key, 16);
    tund_set_key(cnf, tun_ind, 1);
    tun->flags |= TUND_F_RPWD;
    control->c.mt = TUND_CMT_PWDACK;
    control->c.body_len = 0;
    send_control(cnf, tun_ind); 
    return(ex);
#   undef blin_internal_flags
}

static int
generate_new_password(tund_conf *cnf) {
#   define blin_internal_flags (cnf->flags & BLIN_MASK)
    int                   ex = EX_OK;
    static struct timeval tp;
    MD5_CTX               ctx;
    static struct {
        int rn1;
        int t1;
        int rn2;
        int t2;
    }                     buffer;

    if  (gettimeofday(&tp, NULL)) {
        ifBLIN_QW0("gettimeofday");
        ex = EX_OSERR;
        goto out;
    }
    buffer.rn1 = tund_get_rand(cnf);
    buffer.rn2 = tund_get_rand(cnf);
    buffer.t1 = tp.tv_sec;
    buffer.t2 = tp.tv_usec;
    MD5Init(&ctx);
    MD5Update(&ctx, &buffer, sizeof(buffer));
    MD5Final(sKey, &ctx);
out:
    return(ex);
#   undef blin_internal_flags
}

/* Initiates symmetric password change procedure */

static int
change_password(tund_conf *cnf, int tun_ind) {
#   define blin_internal_flags (cnf->flags & BLIN_MASK)
    int          ex = EX_OK;
    tund_tunnel *tun;

    tun = (tund_tunnel *)cnf->stun->buf[tun_ind];
    if  (!!(ex = generate_new_password(cnf))) {
        ifBLIN_QW0("generate_new_password");
        goto out;
    }
    bcopy(sKey, tun->local_key, 16);
    tund_set_key(cnf, tun_ind, 0);
    tun->state = TUND_S_SPAW;
    send_new_password(cnf, tun_ind);
out:
    return(ex);
#   undef blin_internal_flags
}

static int
accept_reset(tund_conf *cnf, int tun_ind) {
#   define blin_internal_flags (cnf->flags & BLIN_MASK)
    tund_CP     *control;
    int          ex = EX_OK;
    tund_tunnel *tun;

    tund_remove_expired_operation(cnf, tun_ind);
    tun = (tund_tunnel *)cnf->stun->buf[tun_ind];
    control = (tund_CP *)packetBuf;
    tun->state = TUND_S_INI;
    tun->flags &= ~TUND_F_RPWD;
    control->c.mt = TUND_CMT_RSTACK;
    control->c.body_len = 0;
    send_control(cnf, tun_ind); 
    if  (!!(ex = change_password(cnf, tun_ind))) {
        ifBLIN_QW0("change_password");
        goto out;
    }
out:
    return(ex);
#   undef blin_internal_flags
}

static int
accept_reset_ack(tund_conf *cnf, int tun_ind) {
#   define blin_internal_flags (cnf->flags & BLIN_MASK)
    int          ex = EX_OK;
    tund_tunnel *tun;

    tun = (tund_tunnel *)cnf->stun->buf[tun_ind];
    tund_remove_expired_operation(cnf, tun_ind);
    tun->state = TUND_S_INI;
    ifBLIN_QX3("tunnel %u now in initial state f=0x%x", tun->tunum, tun->flags);
    /* Now generate and send symmetric password to remote */
    if  (!!(ex = change_password(cnf, tun_ind))) {
        ifBLIN_QW0("change_password");
        goto out;
    }
out:
    return(ex);
#   undef blin_internal_flags
}

static void
accept_npwd_ack(tund_conf *cnf, int tun_ind) {
#   define blin_internal_flags (cnf->flags & BLIN_MASK)
    tund_tunnel *tun;

    tun = (tund_tunnel *)cnf->stun->buf[tun_ind];
    tun->state = TUND_S_UP;
    tund_remove_expired_operation(cnf, tun_ind);
    ifBLIN_QX3("tunnel %u now in operation state, f=0x%x", tun->tunum, tun->flags);
#   undef blin_internal_flags
}

int
tund_proc_cp(tund_conf *cnf, int tun_ind, int len) {
#   define blin_internal_flags (cnf->flags & BLIN_MASK)
    int          ex = EX_OK;
    int          ret;
    tund_tunnel *tun;
    tund_CP     *cp;

    tun = (tund_tunnel *)cnf->stun->buf[tun_ind];
    cp = (tund_CP *)packetBuf;
    ifBLIN_QX5("tund_proc_cp(%d %d)", tun_ind, len);
    ret = control_decrypt(cnf, cp, tun, len);
    if  (ret < 0) {
        ifBLIN_QX3("Dropped packet (illegal RSA encryption) from tun %u l=%d", tun->tunum, len);
        return(ex);
    }
    ifBLIN_QX5("tund_CP type=%d", cp->c.mt);
    switch(cp->c.mt) {
    case TUND_CMT_SPWD:
        accept_new_password(cnf, tun_ind);
        break;
    case TUND_CMT_PWDACK:
        accept_npwd_ack(cnf, tun_ind);
        break;
    case TUND_CMT_RESET:
        if  (!!(ex = accept_reset(cnf, tun_ind))) {
            ifBLIN_QW0("accept_reset");
            goto out;
        }
        break;
    case TUND_CMT_RSTACK:
        if  (!!(ex = accept_reset_ack(cnf, tun_ind))) {
            ifBLIN_QW0("accept_reset_ack");
            goto out;
        }
        break;
    default:
        ifBLIN_QX3("Dropped illegal tund_CP (unknown type): t=%u mt=%d", tun->tunum, cp->c.mt);
    }
out:
    return(ex);
#   undef blin_internal_flags
}
