/*-
 * Copyright Sergey Kosyakov ks@itp.ac.ru 1999
 * Copyright (C)2002..2020 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: tund.h,v 1.21 2020/12/05 17:03:19 babolo Exp $
 */

#ifndef __TUND_H_
#define __TUND_H_

#define TUND_CONFDIR    "/usr/local/etc"
#define TUND_CONFILE    "tund.conf"
#define TUND_SEEDFILE   "tund.seed.rng"
#define TUND_RSAKEYF    "tund.rsa.key"
#define TUND_RSAPUBDB   "tund.hosts.key"
#define TUND_RSAPUBKEYF "rsa_pub.key"
#define TUND_LOCKFILE   "/var/run/tund.pid"
#define TUND_BUFSIZE    65532
#define TUND_PKT_MAXSZ	(sizeof(tund_CP) + 16)
#define TUND_TABLE_SIZE (1<<16)

typedef enum
{ TUND_CF_RULE
, TUND_CF_DVRT
, TUND_CF_HOST
, TUND_CF_ENCR
, TUND_CF_BFIS
, TUND_CF_IDEA
, TUND_CF_RC5
} tund_cf;

typedef struct {                       /* Common use sequence type                   */
    int    length;
    int    maxlen;
    void **buf;
} tund_seq;

typedef struct {
    u_int32_t  flags;
#   define     TUND_FOREG  0x000200
    int        pid_fd;
    char      *bindaddr;
    char      *port;
    char      *pid_nm;
    char      *confdir;
    char      *conf;
    char      *rsak;
    char      *rsap;
    char      *seed;
    tund_seq  *stun;
    in_addr_t  s_addr;
    int        isock;
    in_port_t  s_port;
    u_int16_t  tunTable[TUND_TABLE_SIZE];
} tund_conf;

typedef struct {                                                           /* tund_tunnel parameters */
    int                dp;             /* divert port                                */
    int                socketDivert;   /* divert socket                              */
    struct sockaddr_in l_addr;         /* local IP address                           */
    struct sockaddr_in to;             /* remote host+port address                   */
    u_char             local_key[16];  /* local symmetric encryption key             */
    u_char             remote_key[16]; /* peer's symmetric encryption key            */
    int              (*f_exp)(tund_conf *cnf, int);/* expire callback, arg=tun_ind           */
    RSA               *r_key;          /* remote RSA public key                      */
    void              *localCipherKey; /* symmetric cipher specific key              */
    void              *remoteCipherKey;/* symmetric cipher specific key              */
    const char        *remoteName;     /* remote host name                           */
    time_t             timer;          /* operation expiration timer                 */
    u_int16_t          tunum;          /* tunnel unique number                       */
    u_int8_t           state;
#   define             TUND_S_INI       0      /* initial state                              */
#   define             TUND_S_RSTAW     2      /* waiting for reset acknolegement            */
#   define             TUND_S_SPAW      4      /* symmetric password acknolegment waiting    */
#   define             TUND_S_UP       16      /* tunnel in operation state                  */
    u_int8_t           flags;
#   define             TUND_F_RPWD	2      /* tunnel already has remote symmetric passwd */
#   define             TUND_CIPHER_BF	0x10   /* BlowFish cipher                            */
#   define             TUND_CIPHER_IDEA	0x20   /* IDEA cipher                                */
#   define             TUND_CIPHER_RC5	0x40   /* RC5 cipher                                 */
#   define             TUND_CIPHER_MASK 0xF0
} tund_tunnel;

typedef struct {                       /* POLL callbacks                             */
    int    tun_ind;                    /* tunnel number in extern tund_seq *stun     */
    int  (*f)(tund_conf *cnf, int tun_ind, int socket);
} tund_CBF;

/*** Protocol packets ***/

typedef struct {                       /* Common packet header                       */
    u_char    ver;                     /* protocol version                           */
#   define    TUND_PROTO_VER  1        /* protocol version=1                         */
    u_char    flags;                   /* packet flags                               */
#   define    TUND_PPF_DATA   1        /* data packet                                */
#   define    TUND_PPF_CNT    2        /* control packet                             */
    u_int16_t tunum;                   /* tunnel unique number                       */
} tund_pkthead;

typedef struct {                       /* Control packet header                      */
    u_int32_t rn;                      /* random number                              */
    u_int16_t mt;                      /* message type                               */
#   define    TUND_CMT_SPWD   1        /* new symmetric password                     */
#   define    TUND_CMT_PWDACK 2        /* TUND_CMT_SPWD receiving acknolegment       */
#   define    TUND_CMT_RESET  3        /* start tunnel ini procedure on peer         */
#   define    TUND_CMT_RSTACK 4        /* TUND_CMT_RESET acknolegement               */
    u_int16_t body_len;                /* lenght of body                             */
} tund_ctrlhead;

typedef struct {                       /* Control packet structure                   */
    tund_pkthead  h;
    tund_ctrlhead c;
    u_char        body[TUND_BUFSIZE];
} tund_CP;

typedef struct {                       /* Data packet structure                      */
    tund_pkthead h;
    u_char       body[TUND_BUFSIZE];
} tund_PDP;

typedef struct {                       /* Encrypted data packet structure            */
    tund_pkthead h;
    u_int16_t    rn;                   /* random number                              */
    u_int16_t    body_len;             /* body length                                */
    u_char       body[TUND_BUFSIZE];
} tund_EDP;

int       tund_remove_expired_operation __P((tund_conf *cnf, int tun_ind));
tund_seq *tund_seq_new                  __P((tund_conf *cnf));
int       tund_seq_append               __P((tund_conf *cnf, tund_seq *s, void *item));
int       tund_add_expired_operation    __P(( tund_conf *cnf
                                            , int        tun_ind
                                            , int        t
                                            , int      (*f_exp)(tund_conf *cnf, int)
                                           ));

/* Socket procedures */
int tund_read_pkt __P((tund_conf *cnf, int tun_ind, int s));
int tund_inp_dvrt __P((tund_conf *cnf, int tun_ind, int s));

/* Security functions */
int            tund_init_alarm   __P((tund_conf *cnf));
int            tund_init_secur   __P((tund_conf *cnf));
int            tund_start_sectun __P((tund_conf *cnf, int tun_ind));
int            tund_data_enc     __P((tund_conf *cnf, u_char *buf, int len, int tun_ind));
void           tund_data_dec     __P((tund_conf *cnf, u_char *buf, int len, int tun_ind));
int            tund_proc_cp      __P((tund_conf *cnf, int tun_ind, int len));

/* Utils */
int   tund_get_rand     __P((tund_conf *cnf));
int   tund_init_rng     __P((tund_conf *cnf));
int   tund_ks_send      __P(( tund_conf       *cnf
                            , int              tun_ind
                            , int              s
                            , void            *msg
                            , size_t           len
                            , struct sockaddr *to
                            , int              tolen
                           ));

#endif /* __TUND_H_ */
